import json

from xml.etree.ElementTree import Element
from xml.etree import cElementTree as eTree

from messagebus import (
    ApplicationErrorMessage,
    MessageBus,
    Message,
    DataType,
    DefaultToken,
)
from pos_api import PosService
from product_api import ProductService
from typing import (
    Dict,
    Optional,
    Union,
    List,
)
from .dto import (
    DeliveryInfoDto,
    Order,
    parse_order,
    CustomProperty,
    SaleLineIdentifier,
    OrderState,
)
from .dto.exception import (
    CannotAddProductWithMandatoryCustomizationOnTotal,
    OrderException,
    ProductBarcodeNotFound,
    ProductInRupture,
    UnavailableProductStock,
)

from .error import OrderError
from .token import OrderApiTokens


class OrderService(object):

    def __init__(
        self,                   # type: OrderService
        message_bus,            # type: MessageBus
        pos_id,                 # type: int
        pos_service=None,       # type: Optional[PosService]
        product_service=None,   # type: Optional[ProductService]
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.pos_id = pos_id

        self.component_name = "ORDERMGR{}".format(self.pos_id)
        if pos_service is not None:
            self.pos_service = pos_service
        else:
            self.pos_service = PosService(
                message_bus=message_bus,
                pos_id=pos_id,
            )

        if product_service is not None:
            self.product_service = product_service
        else:
            self.product_service = ProductService(message_bus=message_bus)

    def create_order(
        self,                       # type: OrderService
        sale_type,                  # type: str
        session_id=None,            # type: Optional[str]
        pod_type=None,              # type: Optional[str]
        order_type="SALE",          # type: str
        order_sub_type=None,        # type: Optional[str]
        service_period_id=None,     # type: Optional[int]
        delivery_info=None,         # type: Optional[DeliveryInfoDto]
        additional_info=None,       # type: Optional[str]
        custom_properties=None,     # type: Optional[Dict[str, str]]
        blk_notify=False,           # type: bool
    ):
        # type: (...) -> Element

        if session_id is None:
            session_id = self.pos_service.get_operator_session()

        if pod_type is None:
            pod_type = self.pos_service.get_pod_type()

        request = {
            "blkNotify": blk_notify,
            "posId": self.pos_id,
            "sessionId": session_id,
            "saleType": sale_type,
            "podType": pod_type,
            "orderType": order_type,
        }
        if order_sub_type is not None:
            request["orderSubType"] = order_sub_type
        if additional_info is not None:
            request["additionalInfo"] = additional_info
        if service_period_id is not None:
            request["servicePeriodId"] = service_period_id
        if custom_properties is not None:
            request["customProperties"] = custom_properties
        if delivery_info is not None:
            request["deliveryInfo"] = delivery_info.to_dict()

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_CREATEORDER,
                data_type=DataType.param,
                data=json.dumps(request).encode("utf-8"),
                timeout=600000000,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        return eTree.XML(msg.data)

    def do_sale(
        self,                           # type: OrderService
        item_id=None,                   # type: Optional[str]
        barcode=None,                   # type: Optional[str]
        qty=1,                          # type: float
        apply_default_options=True,     # type: bool
        custom_properties=None,         # type: Optional[Dict[str, str]]
        blk_notify=False,               # type: bool
        sell_ignoring_rupture=False,    # type: bool
    ):
        # type: (...) -> eTree.Element

        if custom_properties is not None:
            custom_properties = json.dumps(custom_properties)

        params = {
            "blkNotify": blk_notify,
            "posId": int(self.pos_id),
            "itemId": item_id,
            "barcode": barcode,
            "qty": qty,
            "applyDefaultOptions": apply_default_options,
            "customProperties": custom_properties,
            "sellIgnoringRupture": sell_ignoring_rupture,
        }
        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_DOSALE,
            data=json.dumps(params),
        )

    def do_option(
        self,                           # type: OrderService
        item_id,                        # type: str
        qty,                            # type: Union[str, int]
        line_number="",                 # type: Union[str, int]
        apply_default_options=False,    # type: bool
        blk_notify=False,               # type: bool
    ):
        # type: (...) -> eTree.Element

        blk_notify = "1" if blk_notify else "0"
        apply_default_options = "1" if apply_default_options else "0"
        params = (blk_notify, self.pos_id, item_id, qty, line_number, "", apply_default_options)
        buffer = "\0".join(str(p) for p in params)

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_DOOPTION,
            data=buffer,
        )

    def do_tender(
        self,                   # type: OrderService
        tender_id,              # type: Union[int, str]
        amount,                 # type: Union[float, str]
        do_not_close=False,     # type: bool
        order_tender_id=None,   # type: Optional[Union[int, str]]
        tip_amount=None,        # type: Optional[float]
        tender_detail=None,     # type: Optional[Union[Dict[str, Union[bool, int, str]], str]]
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "tenderId": int(tender_id),
            "amount": float(amount),
            "doNotClose": do_not_close,
            "orderTenderId": int(order_tender_id) if order_tender_id not in [None, ""] else None,
            "tipAmount": float(tip_amount) if tip_amount not in [None, ""] else None,
            "tenderDetail": json.dumps(tender_detail) if isinstance(tender_detail, dict) else tender_detail,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_DOTENDER,
            data=json.dumps(request),
        )

    def change_modifier(
        self,               # type: OrderService
        line_number,        # type: int
        item_id,            # type: str
        level,              # type: int
        part_code,          # type: str
        new_qty,            # type: Union[str, int]
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        blk_notify = "1" if blk_notify else "0"
        params = (blk_notify, self.pos_id, line_number, item_id, level, part_code, new_qty)
        buffer = "\0".join(str(p) for p in params)

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_CHANGEMODIFIER,
            data=buffer,
        )

    def clear_option(
        self,               # type: OrderService
        line_number,        # type: Union[int, str]
        quantity="",        # type: Union[int, str]
        item_id="",         # type: str
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        params = ("1" if blk_notify else "0", self.pos_id, line_number, item_id, quantity)

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_CLEAROPTION,
            data="\x00".join((str(p) for p in params)),
        )

    def get_order_picture(
        self,               # type: OrderService
        order_id=None,      # type: int
        no_cache=False,     # type: bool
    ):
        # type: (...) -> Optional[Element]

        no_cache = "true" if no_cache else "false"
        data = "{}\0{}\0{}".format(str(self.pos_id), str(order_id or ""), no_cache).encode("utf-8")
        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_ORDERPICT,
                data_type=DataType.param,
                data=data,
                timeout=600000000,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        order_str = msg.data.split(b"\0")[2]
        if not order_str:
            return None

        order_picture = eTree.XML(order_str)
        if order_id and order_picture.tag != "Order":
            order_picture = order_picture.findall(".//Order")[0]

        return order_picture

    def get_order(
        self,           # type: OrderService
        order_id,       # type: int
    ):
        # type: (...) -> Order

        order_xml = self.get_order_picture(order_id=order_id)
        return parse_order(order_xml=order_xml)

    def has_current_order(
        self,               # type: OrderService
        order_picture=None,  # type: Optional[eTree.Element]
    ):
        # type: (...) -> bool

        if order_picture is None:
            order_picture = self.get_order_picture()
        if order_picture is None:
            return False

        state = order_picture.get("state")
        return state in ["IN_PROGRESS", "TOTALED"]

    def void_lines(
        self,                   # type: OrderService
        line_numbers,           # type: List[int]
        blk_notify=False,       # type: bool
        custom_property=None,   # type: Optional[List[Dict[str, str]]]
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "lineNumbers": ([int(line) for line in line_numbers]) or None,
            "customProperties": custom_property or [],
            "blkNotify": blk_notify,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_VOIDLINE,
            data=json.dumps(request),
        )

    def void_order(
        self,                   # type: OrderService
        order_id="",            # type: Union[str, int]
        abandon=False,          # type: bool
        order_status="",        # type: str
        void_reason="",         # type: str
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "orderId": int(order_id) if order_id else None,
            "abandon": abandon,
            "orderStatus": int(order_status) if order_status not in ("", None) else None,
            "voidReason": void_reason if void_reason else None,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_VOIDORDER,
            data=json.dumps(request),
        )

    def store_order(
        self,               # type: OrderService
        blk_notify=False,   # type: str
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_STOREORDER,
            data=json.dumps(request),
        )

    def recall_order(
        self,                       # type: OrderService
        order_id,                   # type: Union[int, str]
        session_id="",              # type: str
        blk_notify=False,           # type: str
        notify_ndiscount=True,      # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "orderId": int(order_id),
            "sessionId": session_id,
            "blkNotify": blk_notify,
            "notifyNdiscount": notify_ndiscount,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_RECALLORDER,
            data=json.dumps(request),
        )

    def pull_orders(
        self,           # type: OrderService
        orders_ids,     # type: List[int]
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "ordersIds": orders_ids,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_PULLORDERS,
            data=json.dumps(request),
        )

    def reopen_order(
        self,               # type: OrderService
        blk_notify=False,   # type: Optional[bool]
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_ORDERREOPEN,
            data=json.dumps(request),
        )

    def total_order(
        self,                   # type: OrderService
        blk_notify=False,       # type: bool
        notify_ndiscount=True,  # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "notifyNdiscount": notify_ndiscount,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_ORDERTOTAL,
            data=json.dumps(request),
        )

    def change_line_quantity(
        self,               # type: OrderService
        line_number,        # type: int
        item_id,            # type: str
        new_quantity,       # type: int
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "lineNumber": line_number,
            "itemId": item_id,
            "newQuantity": new_quantity,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_CHANGEQTY,
            data=json.dumps(request),
        )

    def set_order_custom_property(
        self,               # type: OrderService
        custom_property,    # type: CustomProperty
        order_id=None,      # type: Optional[int]
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> Optional[eTree.Element]

        return self._set_custom_properties(
            custom_properties=[custom_property],
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def set_order_custom_properties(
        self,               # type: OrderService
        custom_properties,  # type: List[CustomProperty]
        order_id=None,      # type: Optional[int]
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        return self._set_custom_properties(
            custom_properties=custom_properties,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def remove_order_custom_property(
        self,               # type: OrderService
        key,                # type: str
        order_id=None,      # type: Optional[int]
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        return self._set_custom_properties(
            custom_properties=[CustomProperty(
                key=key,
                value="clear",
            )],
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def remove_order_custom_properties(
        self,               # type: OrderService
        keys,               # type: List[str]
        order_id=None,      # type: Optional[int]
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        remove_properties = [
            CustomProperty(
                key=key,
                value="clear",
            ) for key in keys
        ]
        return self._set_custom_properties(
            custom_properties=remove_properties,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def set_sale_line_custom_property(
        self,                   # type: OrderService
        custom_property,        # type: CustomProperty
        sale_line_identifier,   # type: SaleLineIdentifier
        order_id=None,          # type: Optional[int]
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        return self._set_custom_properties(
            custom_properties=[custom_property],
            sale_line_identifier=sale_line_identifier,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def set_sale_line_custom_properties(
        self,                   # type: OrderService
        custom_properties,      # type: List[CustomProperty]
        sale_line_identifier,   # type: SaleLineIdentifier
        order_id=None,          # type: Optional[int]
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        return self._set_custom_properties(
            custom_properties=custom_properties,
            sale_line_identifier=sale_line_identifier,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def remove_sale_line_custom_property(
        self,                   # type: OrderService
        key,                    # type: str
        sale_line_identifier,   # type: SaleLineIdentifier
        order_id=None,          # type: Optional[int]
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        return self._set_custom_properties(
            custom_properties=[CustomProperty(
                key=key,
                value="clear",
            )],
            sale_line_identifier=sale_line_identifier,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def remove_sale_line_custom_properties(
        self,                   # type: OrderService
        keys,                   # type: List[str]
        sale_line_identifier,   # type: SaleLineIdentifier
        order_id=None,          # type: Optional[int]
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        remove_properties = [
            CustomProperty(
                key=key,
                value="clear",
            ) for key in keys
        ]
        return self._set_custom_properties(
            custom_properties=remove_properties,
            sale_line_identifier=sale_line_identifier,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def _set_custom_properties(
        self,                       # type: OrderService
        custom_properties,          # type: List[CustomProperty]
        sale_line_identifier=None,  # type: SaleLineIdentifier
        order_id=None,              # type: Optional[int]
        blk_notify=False,           # type: bool
    ):
        # type: (...) -> Optional[eTree.Element]

        properties = {}
        for prop in custom_properties:
            if not isinstance(prop.value, type(u"")):
                if isinstance(prop.value, bytes):
                    prop.value = prop.value.decode("utf-8")
                else:
                    prop.value = str(prop.value)

            properties[prop.key] = prop.value

        request = {
            "customProperties": properties,
            "posId": int(self.pos_id),
            "orderId": order_id,
            "lineNumber": sale_line_identifier.line_number if sale_line_identifier else None,
            "itemId": sale_line_identifier.item_id if sale_line_identifier else None,
            "level": sale_line_identifier.level if sale_line_identifier else None,
            "partCode": sale_line_identifier.part_code if sale_line_identifier else None,
            "blkNotify": blk_notify,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_SETCUSTOMPROPERTIES,
            data=json.dumps(request),
        )

    def list_orders(
        self,                    # type: OrderService
        states=None,             # type: Optional[List[OrderState]]
        pos_id=None,             # type: Optional[str]
        pod_type=None,           # type: Optional[str]
        since=None,              # type: Optional[str]
        until=None,              # type: Optional[str]
        session_id=None,         # type: Optional[str]
        period=None,             # type: Optional[str]
        limit=None,              # type: Optional[str]
        include_property=None,   # type: Optional[str]
        exclude_property=None,   # type: Optional[str]
        property_val=None,       # type: Optional[str]
        descending=False,        # type: Optional[bool]
    ):
        # type: (...) -> List[Dict[str, Union[str, dict, list]]]

        data = {
            "states": [state.value for state in states] if states is not None else [],
            "posId": pos_id,
            "podType": pod_type,
            "since": since,
            "until": until,
            "sessionId": session_id,
            "period": period,
            "limit": limit,
            "includeProperty": include_property,
            "excludeProperty": exclude_property,
            "propertyValue": property_val,
            "descending": descending,
        }

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_LISTORDERS,
                data_type=DataType.json,
                data=json.dumps(data).encode("utf-8"),
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        return self._parse_listed_orders(orders_data=msg.data.decode("utf-8"))

    def get_order_config(
        self,    # type: OrderService
    ):
        # type: (...) -> Dict[str, Union[str, int]]

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_GET_CONFIG,
            )
        )
        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        return json.loads(msg.data.decode("utf-8"))

    @staticmethod
    def _parse_listed_orders(
        orders_data,    # type: str
    ):
        # type: (...) -> List[Dict[str, Union[str, dict, list]]]

        response_params = orders_data.split("\0")
        xml = eTree.XML(response_params[2].encode("utf-8"))
        result = []
        for order in xml.findall("Order"):
            order_dict = dict(order.items())
            order_dict["custom_properties"] = {}
            for order_property in order.findall("CustomOrderProperties/OrderProperty"):
                order_dict["custom_properties"][order_property.get("key")] = order_property.get("value")

            order_dict["tender_history"] = []
            for tender in order.findall("TenderHistory/Tender"):
                order_dict["tender_history"].append(dict(tender.items()))

            result.append(order_dict)

        return result

    @staticmethod
    def _build_order_exception(
        msg,  # type: Message
    ):
        # type: (...) -> OrderException

        try:
            data_text = msg.data.decode("utf-8")
        except UnicodeDecodeError:
            data_text = msg.data.hex()

        parts = data_text.split("\0")
        if len(parts) >= 2:
            error_code = parts[0]
            error_message = parts[1]
            return OrderException(
                code=error_code,
                description=error_message,
            )

        return OrderException(
            code=-1,
            description="Unknown error: {}".format(data_text),
        )

    def _get_order_picture_from_success_message_response(
        self,   # type: OrderService
        token,  # type: int
        data,   # type: str
    ):
        # type: (...) -> eTree.Element

        reply = self._get_success_message_response(
            token=token,
            data=data,
        )

        return eTree.XML(reply.data)

    def _get_success_message_response(
        self,   # type: OrderService
        token,  # type: int
        data,   # type: str
    ):
        # type: (...) -> Message

        reply = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=token,
                data_type=DataType.param,
                data=data.encode("utf-8"),
            ),
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            application_error = ApplicationErrorMessage.build_from_message(message=reply)
            if application_error.error_code == OrderError.PRODUCT_IN_RUPTURE.code:
                raise ProductInRupture(
                    part_code=int(application_error.error_message),
                    part_name=self._get_part_name(application_error.error_message),
                )

            if application_error.error_code == OrderError.UNAVAILABLE_PRODUCT_STOCK.code:
                raise UnavailableProductStock(
                    part_code=int(application_error.error_message),
                    part_name=self._get_part_name(application_error.error_message),
                )
            if application_error.error_code == OrderError.ADD_PRODUCT_WITH_MANDATORY_CUSTOMIZATION_ON_TOTAL.code:
                raise CannotAddProductWithMandatoryCustomizationOnTotal(
                    part_code=int(application_error.error_message),
                    part_name=self._get_part_name(application_error.error_message),
                )
            if application_error.error_code == OrderError.PRODUCT_BARCODE_NOT_FOUND.code:
                raise ProductBarcodeNotFound(
                    part_code=application_error.error_message,
                )

            if application_error.error_code == OrderError.INVALID_CONTEXT.code:
                raise OrderException(
                    code=OrderError.INVALID_CONTEXT.code,
                    description=application_error.error_message
                )

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=reply)

        return reply

    def _get_part_name(
        self,       # type: OrderService
        part_code,  # type: str
    ):
        # type: (...) -> str

        products_name = self.product_service.get_products_name()

        return products_name.get(part_code, part_code)



from xml.etree import cElementTree as eTree
from sysactions import get_user_information

from model.customexception import NotFoundUserError
from thread_local import (
    set_thread_var,
    get_thread_var,
    get_current_model,
)


def get_current_operator_level():
    if get_thread_var("user_level"):
        return get_thread_var("user_level")

    model = get_current_model()
    pos_user_id = int(model.find("Operator").get('id'))
    if not pos_user_id:
        return 0

    user_level = get_user_level(user_id=pos_user_id)
    set_thread_var("user_level", user_level)

    return user_level


def get_user_level(user_id):
    if user_id is None:
        return None

    user_xml_str = get_user_information(user_id)
    if user_xml_str is None:
        raise NotFoundUserError("USER_NOT_FOUND")

    user_xml = eTree.XML(user_xml_str)
    user_element = user_xml.find("user")
    user_level = int(user_element.attrib["Level"])

    return user_level



# emso.py - functions for handling Slovenian Unique Master Citizen Numbers
# coding: utf-8
#
# Copyright (C) 2022 Blaž Bregar
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Enotna matična številka občana (Unique Master Citizen Number).

The EMŠO is used for uniquely identify persons including foreign citizens
living in Slovenia, It is issued by Centralni Register Prebivalstva CRP
(Central Citizen Registry).

The number consists of 13 digits and includes the person's date of birth, a
political region of birth and a unique number that encodes a person's gender
followed by a check digit.

More information:

* https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number
* https://sl.wikipedia.org/wiki/Enotna_matična_številka_občana

>>> validate('0101006500006')
'0101006500006'
>>> validate('0101006500007')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def calc_check_digit(number):
    """Calculate the check digit."""
    weights = (7, 6, 5, 4, 3, 2, 7, 6, 5, 4, 3, 2)
    total = sum(int(n) * w for n, w in zip(number, weights))
    return str(-total % 11 % 10)


def get_birth_date(number):
    """Return date of birth from valid EMŠO."""
    number = compact(number)
    day = int(number[:2])
    month = int(number[2:4])
    year = int(number[4:7])
    if year < 800:
        year += 2000
    else:
        year += 1000
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_gender(number):
    """Get the person's birth gender ('M' or 'F')."""
    number = compact(number)
    if int(number[9:12]) < 500:
        return 'M'
    else:
        return 'F'


def get_region(number):
    """Return (political) region from valid EMŠO."""
    return number[7:9]


def validate(number):
    """Check if the number is a valid EMŠO number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    get_birth_date(number)
    if calc_check_digit(number) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid ID. This checks the length,
    formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



from datetime import date
from typing import Optional

from ._PosState import PosState
from ._User import User


class PosModel(object):
    def __init__(
        self,
        pos_id,                 # type: int
        state,                  # type: PosState
        business_period,        # type: Optional[date]
        logged_user,            # type: Optional[User]
    ):
        self.pos_id = pos_id
        self.state = state
        self.business_period = business_period
        self.logged_user = logged_user



from datetime import time


class AlertsConfigurations(object):

    def __init__(
        self,                        # type: AlertsConfigurations
        time_calling,                # type: int
        configured_alerts_endpoint,  # type: unicode
        maintenance_window_start,    # type: time
        maintenance_window_end,      # type: time
        days_to_keep_data,           # type: int
    ):
        # type: (...) -> None

        self.time_calling = time_calling
        self.configured_alerts_endpoint = configured_alerts_endpoint
        self.maintenance_window_start = maintenance_window_start
        self.maintenance_window_end = maintenance_window_end
        self.days_to_keep_data = days_to_keep_data



from ._GroupedSaleItems import GroupedSaleItems
from ._PreAccountSeatTotals import PreAccountSeatTotals
from ._ProductData import ProductData
from ._ProductType import (
    ProductTypeId,
    ProductTypeName,
)



# regon.py - functions for handling REGON numbers
# coding: utf-8
#
# Copyright (C) 2015 Dariusz Choruzy
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""REGON (Rejestr Gospodarki Narodowej, Polish register of economic units).

The REGON (Rejestr Gospodarki Narodowej) is a statistical identification
number for businesses. National entities are assigned a 9-digit number, while
local units append 5 digits to form a 14-digit number.

More information:

* https://bip.stat.gov.pl/en/regon/
* https://wyszukiwarkaregon.stat.gov.pl/appBIR/index.aspx

>>> validate('192598184')
'192598184'
>>> validate('123456785')
'123456785'
>>> validate('192598183')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('12345678512347')
'12345678512347'
>>> validate('12345678612342')  # first check digit invalid
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('12345678512348')  # last check digit invalid
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


def calc_check_digit(number):
    """Calculate the check digit for organisations. The number passed
    should not have the check digit included."""
    if len(number) == 8:
        weights = (8, 9, 2, 3, 4, 5, 6, 7)
    else:
        weights = (2, 4, 8, 5, 0, 9, 7, 3, 6, 1, 2, 4, 8)
    check = sum(w * int(n) for w, n in zip(weights, number))
    return str(check % 11 % 10)


def validate(number):
    """Check if the number is a valid REGON number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) not in (9, 14):
        raise InvalidLength()
    if number[-1] != calc_check_digit(number[:-1]):
        raise InvalidChecksum()
    if len(number) == 14 and number[8] != calc_check_digit(number[:8]):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid REGON number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-
import os
from threading import Thread

import cfgtools
from msgbus import (
    FM_PARAM,
    MBEasyContext,
    TK_STORECFG_GET,
    TK_SYS_NAK,
)

import boto3
from helper import (
    config_logger,
    import_pydevd,
)
from mbcontextmessagehandler import (
    MbContextMessageBus,
    MbContextMessageHandler,
)
from messagehandlerbuilder import TopicClientMessageHandlerBuilder
from topicclient import SqsConsumer

REQUIRED_SERVICES = "StoreWideConfig"


def main():
    import_pydevd(os.environ["LOADERCFG"], 9152, False)

    config_logger(os.environ["LOADERCFG"], "TopicClient")
    config_logger(os.environ["LOADERCFG"], "TopicClientReceiveThread")

    mb_context = MBEasyContext("TopicClient")
    message_bus = MbContextMessageBus(mbcontext=mb_context)
    message_handler = MbContextMessageHandler(
        message_bus=message_bus,
        service_name="TopicClient",
        service_type="TopicClient",
        required_services=REQUIRED_SERVICES,
        event_handler_builder=None,
    )

    store_code = _read_sw_config(mb_context, "Store.Id")
    context = _read_sw_config(mb_context, "BackOffice.Host")
    config = cfgtools.read(os.environ["LOADERCFG"])
    server_queue_url = config.find_value("TopicClient.ServerQueueUrl")
    aws_access_key = config.find_value("TopicClient.AwsAccessKey")
    aws_secret_key = config.find_value("TopicClient.AwsSecretKey")
    subjects = config.find_values("TopicClient.ClientSubjects")
    queue_url = config.find_value("TopicClient.QueueUrl")
    region_name = config.find_value("TopicClient.RegionName")
    endpoint_url = config.find_value("TopicClient.EndpointUrl", None)
    message_maximum_length = int(config.find_value("TopicClient.MessageMaximumLength", 250))

    queue = boto3.resource(
        "sqs",
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_key,
        region_name=region_name,
        endpoint_url=endpoint_url
    ).Queue(queue_url + store_code)

    sqs_consumer = SqsConsumer(message_bus, queue, store_code)
    t = Thread(target=sqs_consumer.run)
    t.daemon = True
    t.start()

    for subject in subjects:
        message_bus.subscribe(subject)

    message_handler_builder = TopicClientMessageHandlerBuilder(
        server_queue_url=server_queue_url,
        endpoint_url=endpoint_url,
        region_name=region_name,
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_key,
        subjects=subjects,
        store_code=store_code,
        context=context,
        message_maximum_length=message_maximum_length,
    )
    message_handler.event_handler_builder = message_handler_builder
    return message_handler.handle_events()


def _read_sw_config(mb_context, key):
    reply = mb_context.MB_EasySendMessage("StoreWideConfig", token=TK_STORECFG_GET, format=FM_PARAM, data=key)
    if reply.token == TK_SYS_NAK:
        return None
    return reply.data.decode()



"""PEP 656 support.

This module implements logic to detect if the currently running Python is
linked against musl, and what musl version is used.
"""

import functools
import re
import subprocess
import sys
from typing import Iterator, NamedTuple, Optional, Sequence

from ._elffile import ELFFile


class _MuslVersion(NamedTuple):
    major: int
    minor: int


def _parse_musl_version(output: str) -> Optional[_MuslVersion]:
    lines = [n for n in (n.strip() for n in output.splitlines()) if n]
    if len(lines) < 2 or lines[0][:4] != "musl":
        return None
    m = re.match(r"Version (\d+)\.(\d+)", lines[1])
    if not m:
        return None
    return _MuslVersion(major=int(m.group(1)), minor=int(m.group(2)))


@functools.lru_cache
def _get_musl_version(executable: str) -> Optional[_MuslVersion]:
    """Detect currently-running musl runtime version.

    This is done by checking the specified executable's dynamic linking
    information, and invoking the loader to parse its output for a version
    string. If the loader is musl, the output would be something like::

        musl libc (x86_64)
        Version 1.2.2
        Dynamic Program Loader
    """
    try:
        with open(executable, "rb") as f:
            ld = ELFFile(f).interpreter
    except (OSError, TypeError, ValueError):
        return None
    if ld is None or "musl" not in ld:
        return None
    proc = subprocess.run([ld], stderr=subprocess.PIPE, text=True)
    return _parse_musl_version(proc.stderr)


def platform_tags(archs: Sequence[str]) -> Iterator[str]:
    """Generate musllinux tags compatible to the current platform.

    :param archs: Sequence of compatible architectures.
        The first one shall be the closest to the actual architecture and be the part of
        platform tag after the ``linux_`` prefix, e.g. ``x86_64``.
        The ``linux_`` prefix is assumed as a prerequisite for the current platform to
        be musllinux-compatible.

    :returns: An iterator of compatible musllinux tags.
    """
    sys_musl = _get_musl_version(sys.executable)
    if sys_musl is None:  # Python not dynamically linked against musl.
        return
    for arch in archs:
        for minor in range(sys_musl.minor, -1, -1):
            yield f"musllinux_{sys_musl.major}_{minor}_{arch}"


if __name__ == "__main__":  # pragma: no cover
    import sysconfig

    plat = sysconfig.get_platform()
    assert plat.startswith("linux-"), "not linux"

    print("plat:", plat)
    print("musl:", _get_musl_version(sys.executable))
    print("tags:", end=" ")
    for t in platform_tags(re.sub(r"[.-]", "_", plat.split("-", 1)[-1])):
        print(t, end="\n      ")



class ObjectListValueIsNotAList(Exception):

    def __init__(
        self,
        name: str,
    ) -> None:

        self.name = name

    def __str__(self) -> str:
        return f"The objectList {self.name} was not a list"

    def __repr__(self):
        return self.__str__()



from ._TableError import TableError


class OperatorNotLoggedIn(TableError):
    def __init__(
        self,       # type: OperatorNotLoggedIn
        pos_id,     # type: int
    ):
        self.pos_id = pos_id

    def get_extra_data(self):
        return {
            "posId": self.pos_id,
        }



import datetime
from typing import (
    Dict,
    Any,
    List,
)

from nreports_api.shared import JSON
from order_api import OrderService
from nreports.builders import DataBuilder


class DeliveryFeeBuilder(DataBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        pos_id = int(kwargs["posId"])
        since_datetime = datetime.datetime.strptime(kwargs["sinceDate"], "%Y%m%d")
        until_datetime = datetime.datetime.strptime(kwargs["untilDate"], "%Y%m%d")
        until_datetime = until_datetime.replace(hour=23, minute=59, second=59)

        data_to_build["STORE_ID"] = self.store_config_repository.get_store_id()
        data_to_build["PERIOD"] = "{} - {}".format(
            since_datetime.strftime("%d/%m/%Y"),
            until_datetime.strftime("%d/%m/%Y"),
        )
        self._build_body(
            pos_id=pos_id,
            data_to_build=data_to_build,
            since_datetime=since_datetime.isoformat(),
            until_datetime=until_datetime.isoformat(),
        )

        return data_to_build

    def _build_body(
        self,
        pos_id: int,
        data_to_build: Dict[str, JSON],
        since_datetime: str,
        until_datetime: str,
    ) -> None:

        order_service = OrderService(
            message_bus=self._message_bus,
            pos_id=pos_id,
        )

        orders = order_service.list_orders(
            state="PAID",
            since=since_datetime,
            until=until_datetime,
            include_property="LOGISTIC_PARTNER_ID",
            property_val="0",
        )
        delivery_info = self._get_delivery_info(orders=orders)

        data_to_build["deliveryInfo"] = delivery_info
        data_to_build["TOTAL_DELIVERIES_QTY"] = sum(item["DELIVERIES_QTY"] for item in delivery_info)
        data_to_build["TOTAL_DELIVERY_PAID"] = sum(item["DELIVERY_FEE"] for item in delivery_info)

    @staticmethod
    def _get_delivery_info(
        orders: List[Dict[str, JSON]],
    ):
        delivery_info_by_delivery_man = {}
        for order in orders:
            delivery_man = order["custom_properties"].get("LOGISTIC_DELIVERYMAN_DATA", "$NOT_IDENTIFIED")
            delivery_fee_value = float(order["custom_properties"].get("DELIVERY_FEE_VALUE", 0))
            if delivery_man not in delivery_info_by_delivery_man:
                delivery_info_by_delivery_man[delivery_man] = {
                    "DELIVERIES_QTY": 1,
                    "DELIVERYMAN_NAME": delivery_man,
                    "DELIVERY_FEE": delivery_fee_value,
                }
                continue

            delivery_info_by_delivery_man[delivery_man]["DELIVERIES_QTY"] += 1
            delivery_info_by_delivery_man[delivery_man]["DELIVERY_FEE"] += delivery_fee_value

        return delivery_info_by_delivery_man.values()



from ._OrderModifiedProcessor import OrderModifiedProcessor
from ._OrderModifiedIgnoredTypesProcessor import OrderModifiedIgnoredTypesProcessor
from ._GetOrderMaxPrepTimeProcessor import GetOrderMaxPrepTimeProcessor
from ._GetOrderXmlProcessor import GetOrderXmlProcessor
from ._SystemOrderParser import SystemOrderParser
from ._PurgeOrdersProcessor import PurgeOrdersProcessor
from ._GetPrintersInfoProcessor import GetPrintersInfoProcessor
from ._PickListOrdersProcessor import PickListOrdersProcessor
from ._DeviceStatusChangedProcessor import DeviceStatusChangedProcessor
from ._GetCoursesProcessor import GetCoursesProcessor



# Copyright (C) AB Strakt
# See LICENSE for details.

"""
pyOpenSSL - A simple wrapper around the OpenSSL library
"""

from OpenSSL import crypto, SSL
from OpenSSL.version import (
    __author__, __copyright__, __email__, __license__, __summary__, __title__,
    __uri__, __version__,
)


__all__ = [
    "SSL", "crypto",

    "__author__", "__copyright__", "__email__", "__license__", "__summary__",
    "__title__", "__uri__", "__version__",
]
